@@ -112,16 +112,25 @@ register your dynamic context switcher.
 
 """
 
+from __future__ import annotations
+
 import functools
 
+from types import FrameType
+from typing import Any, Dict, Iterable, Optional, Set, Tuple, Union
+
 from coverage import files
-from coverage.misc import contract, _needs_to_implement
+from coverage.misc import _needs_to_implement
+from coverage.types import TArc, TConfigurable, TLineNo, TSourceTokenLines
 
 
 class CoveragePlugin:
     """Base class for coverage.py plug-ins."""
 
-    def file_tracer(self, filename):        # pylint: disable=unused-argument
+    _coverage_plugin_name: str
+    _coverage_enabled: bool
+
+    def file_tracer(self, filename: str) -> Optional[FileTracer]: # pylint: disable=unused-argument
         """Get a :class:`FileTracer` object for a file.
 
         Plug-in type: file tracer.
@@ -161,7 +170,10 @@ class CoveragePlugin:
         """
         return None
 
-    def file_reporter(self, filename):      # pylint: disable=unused-argument
+    def file_reporter(                  # type: ignore[return]
+        self,
+        filename: str,                  # pylint: disable=unused-argument
+    ) -> Union[FileReporter, str]:      # str should be Literal["python"]
         """Get the :class:`FileReporter` class to use for a file.
 
         Plug-in type: file tracer.
@@ -175,7 +187,10 @@ class CoveragePlugin:
         """
         _needs_to_implement(self, "file_reporter")
 
-    def dynamic_context(self, frame):       # pylint: disable=unused-argument
+    def dynamic_context(
+        self,
+        frame: FrameType,               # pylint: disable=unused-argument
+    ) -> Optional[str]:
         """Get the dynamically computed context label for `frame`.
 
         Plug-in type: dynamic context.
@@ -191,7 +206,10 @@ class CoveragePlugin:
         """
         return None
 
-    def find_executable_files(self, src_dir):       # pylint: disable=unused-argument
+    def find_executable_files(
+        self,
+        src_dir: str,                   # pylint: disable=unused-argument
+    ) -> Iterable[str]:
         """Yield all of the executable files in `src_dir`, recursively.
 
         Plug-in type: file tracer.
@@ -206,7 +224,7 @@ class CoveragePlugin:
         """
         return []
 
-    def configure(self, config):
+    def configure(self, config: TConfigurable) -> None:
         """Modify the configuration of coverage.py.
 
         Plug-in type: configurer.
@@ -220,7 +238,7 @@ class CoveragePlugin:
         """
         pass
 
-    def sys_info(self):
+    def sys_info(self) -> Iterable[Tuple[str, Any]]:
         """Get a list of information useful for debugging.
 
         Plug-in type: any.
@@ -234,7 +252,12 @@ class CoveragePlugin:
         return []
 
 
-class FileTracer:
+class CoveragePluginBase:
+    """Plugins produce specialized objects, which point back to the original plugin."""
+    _coverage_plugin: CoveragePlugin
+
+
+class FileTracer(CoveragePluginBase):
     """Support needed for files during the execution phase.
 
     File tracer plug-ins implement subclasses of FileTracer to return from
@@ -251,7 +274,7 @@ class FileTracer:
 
     """
 
-    def source_filename(self):
+    def source_filename(self) -> str:       # type: ignore[return]
         """The source file name for this file.
 
         This may be any file name you like.  A key responsibility of a plug-in
@@ -266,7 +289,7 @@ class FileTracer:
         """
         _needs_to_implement(self, "source_filename")
 
-    def has_dynamic_source_filename(self):
+    def has_dynamic_source_filename(self) -> bool:
         """Does this FileTracer have dynamic source file names?
 
         FileTracers can provide dynamically determined file names by
@@ -284,7 +307,11 @@ class FileTracer:
         """
         return False
 
-    def dynamic_source_filename(self, filename, frame):     # pylint: disable=unused-argument
+    def dynamic_source_filename(
+        self,
+        filename: str,                  # pylint: disable=unused-argument
+        frame: FrameType,               # pylint: disable=unused-argument
+    ) -> Optional[str]:
         """Get a dynamically computed source file name.
 
         Some plug-ins need to compute the source file name dynamically for each
@@ -299,7 +326,7 @@ class FileTracer:
         """
         return None
 
-    def line_number_range(self, frame):
+    def line_number_range(self, frame: FrameType) -> Tuple[TLineNo, TLineNo]:
         """Get the range of source line numbers for a given a call frame.
 
         The call frame is examined, and the source line number in the original
@@ -318,7 +345,7 @@ class FileTracer:
 
 
 @functools.total_ordering
-class FileReporter:
+class FileReporter(CoveragePluginBase):
     """Support needed for files during the analysis and reporting phases.
 
     File tracer plug-ins implement a subclass of `FileReporter`, and return
@@ -331,7 +358,7 @@ class FileReporter:
 
     """
 
-    def __init__(self, filename):
+    def __init__(self, filename: str) -> None:
         """Simple initialization of a `FileReporter`.
 
         The `filename` argument is the path to the file being reported.  This
@@ -341,10 +368,10 @@ class FileReporter:
         """
         self.filename = filename
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "<{0.__class__.__name__} filename={0.filename!r}>".format(self)
 
-    def relative_filename(self):
+    def relative_filename(self) -> str:
         """Get the relative file name for this file.
 
         This file path will be displayed in reports.  The default
@@ -355,8 +382,7 @@ class FileReporter:
         """
         return files.relative_filename(self.filename)
 
-    @contract(returns='unicode')
-    def source(self):
+    def source(self) -> str:
         """Get the source for the file.
 
         Returns a Unicode string.
@@ -366,10 +392,10 @@ class FileReporter:
         as a text file, or if you need other encoding support.
 
         """
-        with open(self.filename, "rb") as f:
-            return f.read().decode("utf-8")
+        with open(self.filename, encoding="utf-8") as f:
+            return f.read()
 
-    def lines(self):
+    def lines(self) -> Set[TLineNo]:    # type: ignore[return]
         """Get the executable lines in this file.
 
         Your plug-in must determine which lines in the file were possibly
@@ -380,7 +406,7 @@ class FileReporter:
         """
         _needs_to_implement(self, "lines")
 
-    def excluded_lines(self):
+    def excluded_lines(self) -> Set[TLineNo]:
         """Get the excluded executable lines in this file.
 
         Your plug-in can use any method it likes to allow the user to exclude
@@ -393,7 +419,7 @@ class FileReporter:
         """
         return set()
 
-    def translate_lines(self, lines):
+    def translate_lines(self, lines: Iterable[TLineNo]) -> Set[TLineNo]:
         """Translate recorded lines into reported lines.
 
         Some file formats will want to report lines slightly differently than
@@ -413,7 +439,7 @@ class FileReporter:
         """
         return set(lines)
 
-    def arcs(self):
+    def arcs(self) -> Set[TArc]:
         """Get the executable arcs in this file.
 
         To support branch coverage, your plug-in needs to be able to indicate
@@ -427,7 +453,7 @@ class FileReporter:
         """
         return set()
 
-    def no_branch_lines(self):
+    def no_branch_lines(self) -> Set[TLineNo]:
         """Get the lines excused from branch coverage in this file.
 
         Your plug-in can use any method it likes to allow the user to exclude
@@ -440,7 +466,7 @@ class FileReporter:
         """
         return set()
 
-    def translate_arcs(self, arcs):
+    def translate_arcs(self, arcs: Iterable[TArc]) -> Set[TArc]:
         """Translate recorded arcs into reported arcs.
 
         Similar to :meth:`translate_lines`, but for arcs.  `arcs` is a set of
@@ -451,9 +477,9 @@ class FileReporter:
         The default implementation returns `arcs` unchanged.
 
         """
-        return arcs
+        return set(arcs)
 
-    def exit_counts(self):
+    def exit_counts(self) -> Dict[TLineNo, int]:
         """Get a count of exits from that each line.
 
         To determine which lines are branches, coverage.py looks for lines that
@@ -466,7 +492,12 @@ class FileReporter:
         """
         return {}
 
-    def missing_arc_description(self, start, end, executed_arcs=None):     # pylint: disable=unused-argument
+    def missing_arc_description(
+        self,
+        start: TLineNo,
+        end: TLineNo,
+        executed_arcs: Optional[Iterable[TArc]] = None,     # pylint: disable=unused-argument
+    ) -> str:
         """Provide an English sentence describing a missing arc.
 
         The `start` and `end` arguments are the line numbers of the missing
@@ -481,7 +512,7 @@ class FileReporter:
         """
         return f"Line {start} didn't jump to line {end}"
 
-    def source_token_lines(self):
+    def source_token_lines(self) -> TSourceTokenLines:
         """Generate a series of tokenized lines, one for each line in `source`.
 
         These tokens are used for syntax-colored reports.
@@ -512,10 +543,11 @@ class FileReporter:
         for line in self.source().splitlines():
             yield [('txt', line)]
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         return isinstance(other, FileReporter) and self.filename == other.filename
 
-    def __lt__(self, other):
+    def __lt__(self, other: Any) -> bool:
         return isinstance(other, FileReporter) and self.filename < other.filename
 
-    __hash__ = None     # This object doesn't need to be hashed.
+    # This object doesn't need to be hashed.
+    __hash__ = None         # type: ignore[assignment]
